//
//  EADSessionController.h
//
//  Copyright (c) 2021 OPTOELECTRONICS CO., LTD. All rights reserved.
//


#import <Foundation/Foundation.h>
#import <ExternalAccessory/ExternalAccessory.h>

@class OPHBluetoothService;

//**************************************************************************************************
// @name OPHSessionControllerDelegate <NSObject>
//**************************************************************************************************


/** OPHSessionController Delegate
 */
@protocol OPHBluetoothServiceDelegate <NSObject>
@optional
/** Notifies that the result has been received.
 *  @param service OPHBluetoothService instance.
 *  @param data The result received.
 */- (void)bluetoothService:(OPHBluetoothService *)service receivedData:(NSData *)data;
@end


//**************************************************************************************************
// @name OPHSessionController
//**************************************************************************************************


/** This core class is in charge of I / O processing, session management, and accessory retention with OPH-5000i.
 *  This class is a Singleton class that only exists in your app.
 */
@interface OPHBluetoothService : NSObject <EAAccessoryDelegate, NSStreamDelegate>

/** Returns an instance of OPHBluetoothService.
 * @return Returns an instance of OPHBluetoothService.
 */
+ (OPHBluetoothService *)sharedController;
/** Initializes the specified EAAccesory and protocolString.
 * @param accessory Specify the accessory.
 * @param protocolString Specifies the OPH-5000i protocol string.
 */
- (void)setupControllerForAccessory:(EAAccessory *)accessory withProtocolString:(NSString *)protocolString;

/** Create an EASession and open NSInputStream and NSOutputStream.
 *
 * 1. Set yourself as a delegate in the currently held EA Accessory.
 * 2. Generate and get an EASession. At the time of generation, setupControllerForAccessory: accessory: withProtocolString
 *   Use the EAAccessory and protocolString specified in.
 *
 * @return Returns YES if a BOOL EASession could be generated. Otherwise, it returns NO.
 */
- (BOOL)openSession;

/** Close the EASession and remove the RunLoop. Release EASession and initialize it with nil.
 */
- (void)closeSession;

/** Writes the specified NSData to the session.
 * @param data Data to send
 */
- (BOOL)writeData:(NSData *)data;

/** Send an ACK.
 */
- (void)writeACK;

/** Send an NAK.
 */
- (void)writeNAK;

/** Connected EA Accessory.
 */
@property (weak, nonatomic, readonly) EAAccessory *accessory;

/** ProtocolString of the connected device.
 */
@property (nonatomic, readonly) NSString *protocolString;

/** Delegate that will be executed after receiving data.
 */
@property (nonatomic, weak) id<OPHBluetoothServiceDelegate> delegate;


@end
